@@ -20,6 +20,7 @@
 
 #include "handler.h"
 #include "connectioneditordialog.h"
+#include "configuration.h"
 #include "uiutils.h"
 #include "debug.h"
 
@@ -57,13 +58,14 @@
 #include <KService>
 #include <KServiceTypeTrader>
 #include <KWindowSystem>
-#include <KIconLoader>
 #include <KWallet>
 
 #define AGENT_SERVICE "org.kde.kded5"
 #define AGENT_PATH "/modules/networkmanagement"
 #define AGENT_IFACE "org.kde.plasmanetworkmanagement"
 
+// 10 seconds
+#define NM_REQUESTSCAN_LIMIT_RATE 10000
 
 Handler::Handler(QObject *parent)
     : QObject(parent)
@@ -83,9 +85,23 @@ Handler::Handler(QObject *parent)
                                             QStringLiteral("secretsError"),
                                             this, SLOT(secretAgentError(QString, QString)));
 
-    // Interval (in ms) between attempts to scan for wifi networks
-    m_wirelessScanRetryTimer.setInterval(2000);
-    m_wirelessScanRetryTimer.setSingleShot(true);
+
+    if (!Configuration::hotspotConnectionPath().isEmpty()) {
+        NetworkManager::ActiveConnection::Ptr hotspot = NetworkManager::findActiveConnection(Configuration::hotspotConnectionPath());
+        if (!hotspot) {
+            Configuration::setHotspotConnectionPath(QString());
+            Q_EMIT hotspotDisabled();
+        }
+    }
+
+    m_hotspotSupported = checkHotspotSupported();
+
+    if (NetworkManager::checkVersion(1, 16, 0)) {
+        connect(NetworkManager::notifier(), &NetworkManager::Notifier::primaryConnectionTypeChanged, [this] () {
+            m_hotspotSupported = checkHotspotSupported();
+            Q_EMIT hotspotSupportedChanged(m_hotspotSupported);
+        });
+    }
 }
 
 Handler::~Handler()
@@ -128,7 +144,7 @@ void Handler::activateConnection(const QString& connection, const QString& devic
                 notification->setComponentName("networkmanagement");
                 notification->setTitle(con->name());
                 notification->setText(i18n("Missing VPN plugin"));
-                notification->setPixmap(QIcon::fromTheme("dialog-warning").pixmap(KIconLoader::SizeHuge));
+                notification->setIconName(QStringLiteral("dialog-warning"));
                 notification->sendEvent();
                 return;
             }
@@ -496,6 +512,30 @@ void Handler::requestScan(const QString &interface)
                     continue;
                 }
 
+                if (!checkRequestScanRateLimit(wifiDevice)) {
+                    QDateTime now = QDateTime::currentDateTime();
+                    // for NM < 1.12, lastScan is not available
+                    QDateTime lastScan = wifiDevice->lastScan();
+                    QDateTime lastRequestScan = wifiDevice->lastRequestScan();
+                    // Compute the next time we can run a scan
+                    int timeout = NM_REQUESTSCAN_LIMIT_RATE;
+                    if (lastScan.isValid() && lastScan.msecsTo(now) < NM_REQUESTSCAN_LIMIT_RATE) {
+                        timeout = NM_REQUESTSCAN_LIMIT_RATE - lastScan.msecsTo(now);
+                    } else if (lastRequestScan.isValid() && lastRequestScan.msecsTo(now) < NM_REQUESTSCAN_LIMIT_RATE) {
+                        timeout = NM_REQUESTSCAN_LIMIT_RATE - lastRequestScan.msecsTo(now);
+                    }
+                    qCDebug(PLASMA_NM) << "Rescheduling a request scan for" << wifiDevice->interfaceName() << "in" << timeout;
+                    scheduleRequestScan(wifiDevice->interfaceName(), timeout);
+
+                    if (!interface.isEmpty()) {
+                        return;
+                    }
+                    continue;
+                } else if (m_wirelessScanRetryTimer.contains(interface)){
+                    m_wirelessScanRetryTimer.value(interface)->stop();
+                    delete m_wirelessScanRetryTimer.take(interface);
+                }
+
                 qCDebug(PLASMA_NM) << "Requesting wifi scan on device" << wifiDevice->interfaceName();
                 QDBusPendingReply<> reply = wifiDevice->requestScan();
                 QDBusPendingCallWatcher *watcher = new QDBusPendingCallWatcher(reply, this);
@@ -507,21 +547,193 @@ void Handler::requestScan(const QString &interface)
     }
 }
 
-void Handler::scanRequestFailed(const QString &interface)
+void Handler::createHotspot()
 {
-    if (m_wirelessScanRetryTimer.isActive()) {
+    bool foundInactive = false;
+    bool useApMode = false;
+    NetworkManager::WirelessDevice::Ptr wifiDev;
+
+    NetworkManager::ConnectionSettings::Ptr connectionSettings;
+    connectionSettings = NetworkManager::ConnectionSettings::Ptr(new NetworkManager::ConnectionSettings(NetworkManager::ConnectionSettings::Wireless));
+
+    NetworkManager::WirelessSetting::Ptr wifiSetting = connectionSettings->setting(NetworkManager::Setting::Wireless).dynamicCast<NetworkManager::WirelessSetting>();
+    wifiSetting->setMode(NetworkManager::WirelessSetting::Adhoc);
+    wifiSetting->setSsid(Configuration::hotspotName().toUtf8());
+
+    for (const NetworkManager::Device::Ptr &device : NetworkManager::networkInterfaces()) {
+        if (device->type() == NetworkManager::Device::Wifi) {
+            wifiDev = device.objectCast<NetworkManager::WirelessDevice>();
+            if (wifiDev) {
+                if (!wifiDev->isActive()) {
+                    foundInactive = true;
+                } else {
+                    // Prefer previous device if it was inactive
+                    if (foundInactive) {
+                        break;
+                    }
+                }
+
+                if (wifiDev->wirelessCapabilities().testFlag(NetworkManager::WirelessDevice::ApCap)) {
+                    useApMode = true;
+                }
+
+                // We prefer inactive wireless card with AP capabilities
+                if (foundInactive && useApMode) {
+                    break;
+                }
+            }
+        }
+    }
+
+    if (!wifiDev) {
+        qCWarning(PLASMA_NM) << "Failed to create hotspot: missing wireless device";
         return;
     }
-    qCDebug(PLASMA_NM) << "Trying soon a new scan on" << interface;
 
-    emit wirelessScanTimerEnabled(false);
+    wifiSetting->setInitialized(true);
+    wifiSetting->setMode(useApMode ? NetworkManager::WirelessSetting::Ap :NetworkManager::WirelessSetting::Adhoc);
 
-    auto retryAction = [this,interface]() {
-        m_wirelessScanRetryTimer.disconnect();
-        requestScan(interface);
-    };
-    connect(&m_wirelessScanRetryTimer, &QTimer::timeout, this, retryAction);
-    m_wirelessScanRetryTimer.start();
+    if (!Configuration::hotspotPassword().isEmpty()) {
+        NetworkManager::WirelessSecuritySetting::Ptr wifiSecurity = connectionSettings->setting(NetworkManager::Setting::WirelessSecurity).dynamicCast<NetworkManager::WirelessSecuritySetting>();
+        wifiSecurity->setInitialized(true);
+
+        if (useApMode) {
+            // Use WPA2
+            wifiSecurity->setKeyMgmt(NetworkManager::WirelessSecuritySetting::WpaPsk);
+            wifiSecurity->setPsk(Configuration::hotspotPassword());
+            wifiSecurity->setPskFlags(NetworkManager::Setting::AgentOwned);
+        } else {
+            // Use WEP
+            wifiSecurity->setKeyMgmt(NetworkManager::WirelessSecuritySetting::Wep);
+            wifiSecurity->setWepKeyType(NetworkManager::WirelessSecuritySetting::Passphrase);
+            wifiSecurity->setWepTxKeyindex(0);
+            wifiSecurity->setWepKey0(Configuration::hotspotPassword());
+            wifiSecurity->setWepKeyFlags(NetworkManager::Setting::AgentOwned);
+            wifiSecurity->setAuthAlg(NetworkManager::WirelessSecuritySetting::Open);
+        }
+    }
+
+    NetworkManager::Ipv4Setting::Ptr ipv4Setting = connectionSettings->setting(NetworkManager::Setting::Ipv4).dynamicCast<NetworkManager::Ipv4Setting>();
+    ipv4Setting->setMethod(NetworkManager::Ipv4Setting::Shared);
+    ipv4Setting->setInitialized(true);
+
+    connectionSettings->setId(Configuration::hotspotName());
+    connectionSettings->setAutoconnect(false);
+    connectionSettings->setUuid(NetworkManager::ConnectionSettings::createNewUuid());
+
+    const QVariantMap options = { {QLatin1String("persist"), QLatin1String("volatile")} };
+
+    QDBusPendingReply<QDBusObjectPath, QDBusObjectPath, QVariantMap> reply = NetworkManager::addAndActivateConnection2(connectionSettings->toMap(), wifiDev->uni(), QString(), options);
+    QDBusPendingCallWatcher *watcher = new QDBusPendingCallWatcher(reply, this);
+    watcher->setProperty("action", Handler::CreateHotspot);
+    watcher->setProperty("connection", Configuration::hotspotName());
+    connect(watcher, &QDBusPendingCallWatcher::finished, this, &Handler::replyFinished);
+    connect(watcher, &QDBusPendingCallWatcher::finished, this, QOverload<QDBusPendingCallWatcher *>::of(&Handler::hotspotCreated));
+}
+
+void Handler::stopHotspot()
+{
+    const QString activeConnectionPath = Configuration::hotspotConnectionPath();
+
+    if (activeConnectionPath.isEmpty()) {
+        return;
+    }
+
+    NetworkManager::ActiveConnection::Ptr hotspot = NetworkManager::findActiveConnection(activeConnectionPath);
+
+    if (!hotspot) {
+        return;
+    }
+
+    NetworkManager::deactivateConnection(activeConnectionPath);
+    Configuration::setHotspotConnectionPath(QString());
+
+    Q_EMIT hotspotDisabled();
+}
+
+bool Handler::checkRequestScanRateLimit(const NetworkManager::WirelessDevice::Ptr &wifiDevice)
+{
+    QDateTime now = QDateTime::currentDateTime();
+    QDateTime lastScan = wifiDevice->lastScan();
+    QDateTime lastRequestScan = wifiDevice->lastRequestScan();
+
+    // if the last scan finished within the last 10 seconds
+    bool ret = lastScan.isValid() && lastScan.msecsTo(now) < NM_REQUESTSCAN_LIMIT_RATE;
+    // or if the last Request was sent within the last 10 seconds
+    ret |= lastRequestScan.isValid() && lastRequestScan.msecsTo(now) < NM_REQUESTSCAN_LIMIT_RATE;
+    // skip the request scan
+    if (ret) {
+        qCDebug(PLASMA_NM) << "Last scan finished " << lastScan.msecsTo(now) << "ms ago and last request scan was sent "
+                           << lastRequestScan.msecsTo(now) << "ms ago, Skipping scanning interface:" << wifiDevice->interfaceName();
+        return false;
+    }
+    return true;
+}
+
+bool Handler::checkHotspotSupported()
+{
+    if (NetworkManager::checkVersion(1, 16, 0)) {
+        bool unusedWifiFound = false;
+        bool wifiFound = false;
+
+        for (const NetworkManager::Device::Ptr &device : NetworkManager::networkInterfaces()) {
+            if (device->type() == NetworkManager::Device::Wifi) {
+                wifiFound = true;
+
+                NetworkManager::WirelessDevice::Ptr wifiDev = device.objectCast<NetworkManager::WirelessDevice>();
+                if (wifiDev && !wifiDev->isActive()) {
+                    unusedWifiFound = true;
+                }
+            }
+        }
+
+
+        if (!wifiFound) {
+            return false;
+        }
+
+        if (unusedWifiFound) {
+            return true;
+        }
+
+        // Check if the primary connection which is used for internet connectivity is not using WiFi
+        if (NetworkManager::primaryConnectionType() != NetworkManager::ConnectionSettings::Wireless) {
+            return true;
+        }
+    }
+
+    return false;
+}
+
+void Handler::scheduleRequestScan(const QString &interface, int timeout)
+{
+    QTimer *timer;
+    if (!m_wirelessScanRetryTimer.contains(interface)) {
+        // create a timer for the interface
+        timer = new QTimer();
+        timer->setSingleShot(true);
+        m_wirelessScanRetryTimer.insert(interface, timer);
+        auto retryAction = [this, interface]() {
+            requestScan(interface);
+        };
+        connect(timer, &QTimer::timeout, this, retryAction);
+    } else {
+        // set the new value for an existing timer
+        timer = m_wirelessScanRetryTimer.value(interface);
+        if (timer->isActive()) {
+            timer->stop();
+        }
+    }
+
+    // +1 ms is added to avoid having the scan being rejetted by nm
+    // because it is run at the exact last millisecond of the requestScan threshold
+    timer->setInterval(timeout + 1);
+    timer->start();
+}
+
+void Handler::scanRequestFailed(const QString &interface)
+{
+    scheduleRequestScan(interface, 2000);
 }
 
 void Handler::initKdedModule()
@@ -579,6 +791,10 @@ void Handler::replyFinished(QDBusPendingCallWatcher * watcher)
                 scanRequestFailed(interface);
                 break;
             }
+            case Handler::CreateHotspot:
+                notification = new KNotification("FailedToCreateHotspot", KNotification::CloseOnTimeout, this);
+                notification->setTitle(i18n("Failed to create hotspot %1", watcher->property("connection").toString()));
+                break;
             default:
                 break;
         }
@@ -586,7 +802,7 @@ void Handler::replyFinished(QDBusPendingCallWatcher * watcher)
         if (notification) {
             notification->setComponentName("networkmanagement");
             notification->setText(error);
-            notification->setPixmap(QIcon::fromTheme("dialog-warning").pixmap(KIconLoader::SizeHuge));
+            notification->setIconName(QStringLiteral("dialog-warning"));
             notification->sendEvent();
         }
     } else {
@@ -608,7 +824,6 @@ void Handler::replyFinished(QDBusPendingCallWatcher * watcher)
                 break;
             case Handler::RequestScan:
                 qCDebug(PLASMA_NM) << "Wireless scan on" << watcher->property("interface").toString() << "succeeded";
-                emit wirelessScanTimerEnabled(true);
                 break;
             default:
                 break;
@@ -617,7 +832,7 @@ void Handler::replyFinished(QDBusPendingCallWatcher * watcher)
         if (notification) {
             notification->setComponentName("networkmanagement");
             notification->setTitle(watcher->property("connection").toString());
-            notification->setPixmap(QIcon::fromTheme("dialog-information").pixmap(KIconLoader::SizeHuge));
+            notification->setIconName(QStringLiteral("dialog-information"));
             notification->sendEvent();
         }
     }
@@ -625,6 +840,36 @@ void Handler::replyFinished(QDBusPendingCallWatcher * watcher)
     watcher->deleteLater();
 }
 
+void Handler::hotspotCreated(QDBusPendingCallWatcher *watcher)
+{
+    QDBusPendingReply<QDBusObjectPath, QDBusObjectPath, QVariantMap> reply = *watcher;
+
+    if (!reply.isError() && reply.isValid()) {
+        const QString activeConnectionPath = reply.argumentAt(1).value<QDBusObjectPath>().path();
+
+        if (activeConnectionPath.isEmpty()) {
+            return;
+        }
+
+        Configuration::setHotspotConnectionPath(activeConnectionPath);
+
+        NetworkManager::ActiveConnection::Ptr hotspot = NetworkManager::findActiveConnection(activeConnectionPath);
+
+        if (!hotspot) {
+            return;
+        }
+
+        connect(hotspot.data(), &NetworkManager::ActiveConnection::stateChanged, [=] (NetworkManager::ActiveConnection::State state) {
+            if (state > NetworkManager::ActiveConnection::Activated) {
+                Configuration::setHotspotConnectionPath(QString());
+                Q_EMIT hotspotDisabled();
+            }
+        });
+
+        Q_EMIT hotspotCreated();
+    }
+}
+
 #if WITH_MODEMMANAGER_SUPPORT
 void Handler::unlockRequiredChanged(MMModemLock modemLock)
 {
